unit SimpSock;

interface

uses
  Windows, Messages, SysUtils, WinSock, IndexList, {$IFDEF MINIMAL}MiniClasses{$ELSE}Classes{$ENDIF};

type
  {     }
  
  ESocketException = class(Exception);
  //  

  TSocketErrorType = (seOpen, seClose, seRead, seWrite, seAccept, seGeneral);
  //  ,     

  TSocketEvent = procedure(Sender: TObject) of object;
  //  ,   TNotifyEvent

  TSocketErrorEvent = procedure(Sender: TObject; const Error: TSocketErrorType; const Code: Integer; var Handled: Boolean) of object;
  //  ,         
  // , Handled      TRUE

  TAvailableEvent = procedure(Sender: TObject; const Data: Pointer; const Size: Integer) of object;
  //      
  //      (   )
  //  Data ,     
  //     

  TPongEvent = procedure(Sender: TObject; const PingTime: Cardinal) of object;
  //    Ping


  TBlockBuffer = record
    Data: Pointer;
    Size: Integer;
  end;
  //        

  TWaitAction = (waOpen, waSend);
  //     

{
      
   ,        
       -.   
        , 
     WinSock,    
     .

      ,    
      ,     .
}
  TCustomSocket = class(TObject)
  protected
    FActive: Boolean; //  
    FOnError: TSocketErrorEvent; //  
    FSocket: TSocket; //  
    FAddress: TSockAddr; // IP-  
    FSocketLock: TCriticalSection; //     
    FBlockMode: Boolean;
    FBlockBuffer, FSendBuffer: TBlockBuffer;
    FOnIncoming: TAvailableEvent; //   ;)
    FOnPong: TPongEvent; //   Ping
    FMaxWSAPacketSize: Word; //     IP  
    FMaxBlockSize: Integer; //         
    FTag: Integer;
    procedure   RaiseError(const Error: TSocketErrorType; const Code: Integer); //          (    )
    procedure   DataReceive; //        
    function    Min(const Arg1, Arg2: Integer): Integer; // no comment
    procedure   SetActive(const Value: Boolean); //  
    procedure   SetMaxBlockSize(const Value: Integer);
    procedure   SetMaxWSAPacketSize(const Value: Word);
    procedure   ProcessMessages; virtual; abstract;
    procedure   SetBlockMode(const Value: Boolean);
    function    GetSendBufferLength: Integer;
    function    FindBlockBegin(const DataArray: Pointer; const DataSize: Integer): Pointer; //      (   -)
    procedure   ProcessSendBuffer; virtual; //          WinSock,      
    function    TrySend(const Data: Pointer; const Size: Integer): Integer; //      FMaxWSAPacketSize     WinSock,      Data
    procedure   SafeSend(const Data: Pointer; const Size: Integer); //   /  /    (     ProcessSendBuffer)
    function    Pong(const PingParam: Cardinal): Boolean;
    function    QueryInfo: Boolean;
    function    SendInfo: Boolean;
    procedure   OnNullPacketReceived; virtual;
    procedure   OnDataReceived(const Data: Pointer; const Size: Integer); //     ,      
  public
    constructor Create;
    destructor  Destroy; override;
    procedure   Open; virtual;
    procedure   Close; virtual;
    function    Send(const Data: Pointer; const Size: Integer): Boolean; virtual; //    
    function    WaitFor(const Action: TWaitAction; const TimeOut: Cardinal = 0): Boolean; virtual; //    

    function    SendNullPacket: Boolean;
    function    Ping: Boolean;

    property    Socket: TSocket read FSocket;
    property    Active: Boolean read FActive write SetActive default false;
    property    MaxWSAPacketSize: Word read FMaxWSAPacketSize write SetMaxWSAPacketSize default 4096;

    property    OnError: TSocketErrorEvent read FOnError write FOnError;
    property    OnIncoming: TAvailableEvent read FOnIncoming write FOnIncoming; //    
    property    OnPong: TPongEvent read FOnPong write FOnPong;

    property    BlockMode: Boolean read FBlockMode write SetBlockMode default false; //          
    property    MaxBlockSize: Integer read FMaxBlockSize write SetMaxBlockSize default 10*1024*1024;
    property    Tag: Integer read FTag write FTag default 0;
    property    SendBufferLength: Integer read GetSendBufferLength;
  end;

type
  {      }

{
      ,  .   
    .
}
  TSimpleClientSocket = class(TCustomSocket)
  private
    FPort: Word; // ,    
    FHost: String; // /  
    FWndHandle: HWND; //   ,   
    FOnConnect, FOnDisconnect: TSocketEvent; //  
    procedure   OnQueryEndSession(var Msg: TMessage); message WM_QUERYENDSESSION; //      
    procedure   WindowProc(var Msg: TMessage); //    
  protected
    procedure   FreeResources; //       
    function    LookupName(const Name: String): TInAddr; //    DNS
    procedure   SetHost(const Value: String); //  FHost
    procedure   SetPort(const Value: Word); //  FPort

    property    WndHandle: HWND read FWndHandle;
  public
    constructor Create;
    destructor  Destroy; override;
    procedure   Open; override;
    procedure   Close; override;
    procedure   ProcessMessages; override;
  published
    property    Host: String read FHost write SetHost;
    property    Port: Word read FPort write SetPort default 0;
    property    Tag;
    property    MaxWSAPacketSize;

    property    OnIncoming; //   
    property    BlockMode;
    property    MaxBlockSize;

    property    OnConnect: TSocketEvent read FOnConnect write FOnConnect;
    property    OnDisconnect: TSocketEvent read FOnDisconnect write FOnDisconnect;
    property    OnError; //   
    property    OnPong;
  end;

type
  {      }
  
  TSimpleServerSocket = class;
  TServerClientSocket = class;

  TAcceptRequestEvent = procedure(Sender: TObject; var AllowAccept: Boolean) of object;
  //      ( ).    
  //     .

  TSrvSocketEvent = procedure(Sender: TObject; const ClientEntry: TServerClientSocket) of object;
  //  ,   TNotifyEvent

  TSrvSocketErrorEvent = procedure(Sender: TObject; const ClientEntry: TServerClientSocket; const Error: TSocketErrorType; const Code: Integer; var Handled: Boolean) of object;
  //  ,         
  // , Handled      TRUE

  TSrvAvailableEvent = procedure(Sender: TObject; const ClientEntry: TServerClientSocket; const Data: Pointer; const Size: Integer) of object;
  //      
  //      (   )
  //  Data ,     
  //     

  TSrvAcceptRequestEvent = procedure(Sender: TObject; const ClientEntry: TServerClientSocket; var AllowAccept: Boolean) of object;
  //      ( ).    
  //     .

  TSrvPongEvent = procedure(Sender: TObject; const ClientEntry: TServerClientSocket; const PingTime: Cardinal) of object;
  //    Ping

{
  ,        . 
    ,       .    
  ,   ,        .
}
  TServerClientSocket = class(TCustomSocket)
  private
    FServerSocket: TSimpleServerSocket; //   
    FOnAcceptRequest: TAcceptRequestEvent;
  protected
    FLocalAddress: TSockAddr; // IP-  
    procedure   ProcessMessages; override;
    procedure   FreeResources;

    property    MaxWSAPacketSize; //     / 
    property    BlockMode;
    property    MaxBlockSize;
    property    OnIncoming;
    property    OnAcceptRequest: TAcceptRequestEvent read FOnAcceptRequest write FOnAcceptRequest;
    property    OnPong;
  public
    constructor Create(const AServerSocket: TSimpleServerSocket);
    destructor  Destroy; override;

    procedure   Open; override; //     
    procedure   Close; override; //     ,  Drop!
    procedure   Drop;

    property    Address: TSockAddr read FAddress;
    property    LocalAddress: TSockAddr read FLocalAddress;
    property    ServerSocket: TSimpleServerSocket read FServerSocket;
  end;

{
   .  .   -   
         
  TServerClientSocket,         
   .
}

  { TSimpleServerSocket }

  TSimpleServerSocket = class(TObject)
  private
    FActive: Boolean; //  
    FWndHandle: HWND; //   ,   
    FSocket: TSocket; //   
    FAddress: TSockAddr; // IP-  
    FPort: Word;  //   
    FSocketLock: TCriticalSection; //      
    FClientSockets: TIndexList; //    TServerClientSocket
    FOnIncoming: TSrvAvailableEvent; //   
    FOnOpen, FOnClose: TSocketEvent; //  
    FOnClientConnected, FOnClientDisconnected: TSrvSocketEvent; //   
    FOnClientError: TSrvSocketErrorEvent;
    FOnAcceptRequest: TSrvAcceptRequestEvent;
    FOnClientPong: TSrvPongEvent;
    FMaxWSAPacketSize: Word; //   
    FMaxBlockSize: Integer; //         
    FBlockMode: Boolean;
    FTag: Integer;
    procedure   OnQueryEndSession(var Msg: TMessage); message WM_QUERYENDSESSION; //      
    procedure   SetBlockMode(const Value: Boolean);
    procedure   WindowProc(var Msg: TMessage); //    
    procedure   OnClientEntryError(Sender: TObject; const Error: TSocketErrorType; const Code: Integer; var Handled: Boolean);
    procedure   OnClientEntryIncoming(Sender: TObject; const Data: Pointer; const Size: Integer);
    procedure   OnClientAcceptRequest(Sender: TObject; var AllowAccept: Boolean);
    procedure   OnClientEntryPong(Sender: TObject; const PingTime: Cardinal);
  protected
    FOnError: TSocketErrorEvent; //  
    procedure   FreeResources; //        
    procedure   RaiseError(const Sender: TObject; const Error: TSocketErrorType; const Code: Integer); //          (    )
    procedure   SetActive(const Value: Boolean); //  FActive
    procedure   SetPort(const Value: Word); //  
    function    GetClient(const Index: Integer): TServerClientSocket; //     
    function    GetClientCount: Integer; //   
    function    GetClientBySocket(const SocketToFind: TSocket): TServerClientSocket; //      
    function    GetSendBufferLength: Integer;
    procedure   SetMaxBlockSize(const Value: Integer);
    procedure   SetMaxWSAPacketSize(const Value: Word);
    procedure   SendDropMessage(const Socket: TSocket);

    property    WndHandle: HWND read FWndHandle;
  public
    constructor Create;
    destructor  Destroy; override;
    procedure   Open;
    procedure   Close;
    procedure   ProcessMessages;
    function    FindClientSocket(const SocketToFind: TSocket): Integer; //       
    procedure   DropClientSocket(const Index: Integer); //    
    function    GetNextClientIndex(const CurIndex: Integer): Integer; //      (     ClientCount-1  !)

    procedure   SendNullPacketAll;
    procedure   PingAll;

    property    ClientsBySocket[const SocketToFind: TSocket]: TServerClientSocket read GetClientBySocket; //        
    property    Clients[const Index: Integer]: TServerClientSocket read GetClient; //        
    property    ClientCount: Integer read GetClientCount; //   
    property    Active: Boolean read FActive write SetActive default false;
    property    SendBufferLength: Integer read GetSendBufferLength;
  published
    property    Port: Word read FPort write SetPort default 0;
    property    Socket: TSocket read FSocket;
    property    MaxWSAPacketSize: Word read FMaxWSAPacketSize write SetMaxWSAPacketSize default 4096;
    property    MaxBlockSize: Integer read FMaxBlockSize write SetMaxBlockSize default 10*1024*1024;
    property    BlockMode: Boolean read FBlockMode write SetBlockMode default false; //          
    property    Tag: Integer read FTag write FTag;

    property    OnIncoming: TSrvAvailableEvent read FOnIncoming write FOnIncoming;
    property    OnOpen: TSocketEvent read FOnOpen write FOnOpen;
    property    OnClose: TSocketEvent read FOnClose write FOnClose;
    property    OnError: TSocketErrorEvent read FOnError write FOnError;
    property    OnClientConnected: TSrvSocketEvent read FOnClientConnected write FOnClientConnected;
    property    OnClientDisconnected: TSrvSocketEvent read FOnClientDisconnected write FOnClientDisconnected;
    property    OnClientError: TSrvSocketErrorEvent read FOnClientError write FOnClientError;
    property    OnClientAccept: TSrvAcceptRequestEvent read FOnAcceptRequest write FOnAcceptRequest;
    property    OnClientPong: TSrvPongEvent read FOnClientPong write FOnClientPong;
  end;

type
  TSimpleSocketInfo = packed record
    Size: Integer;
    Version: DWord;
    Comment: string[100];
  end;
  PSimpleSocketInfo = ^TSimpleSocketInfo;
  //     

const
  SocketErrorsText: array [TSocketErrorType] of String = ('Open', 'Close', 'Read', 'Write', 'Accept', 'General');
  //    TSocketErrorType

  CurVersion: DWord = $00000100;
  //   

type
  TIP = packed record
    case Byte of
      0: (Long: Integer);
      1: (b1, b2, b3, b4: Byte);
      2: (c1, c2, c3, c4: Char);
      3: (Bytes: array [1..4] of Byte);
      4: (Chars: array [1..4] of Char);
  end;

  function  SockAddrToIP(const Address: TSockAddr): TIP;
  function  IPToStr(const IP: TIP): String;
  function  StrToIP(const Address: String): TIP;

implementation

uses
  {$IFDEF FPC}
  WndUtil,
  {$ENDIF}
  RTLConsts;
  
/////////////////////////// WinSock2 part //////////////////////////////////
type
  TWSABuf = packed record
    Len: U_LONG;
    Buf: Pointer;
  end;
  PWSABuf = ^TWSABuf;
  TServiceType = Integer;
  TFlowSpec = packed record
    TokenRate,
    TokenBucketSize,
    PeakBandwidth,
    Latency,
    DelayVariation: Integer;
    ServiceType: TServiceType;
    MaxSduSize, MinimumPolicedSize: Integer;
  end;
  TQOS = packed record
    SendingFlowspec: TFlowSpec;
    ReceivingFlowspec: TFlowSpec;
    ProviderSpecific: TWSABuf;
  end;
  PQOS = ^TQOS;
  TGroup = Integer;
  PGroup = ^TGroup;

const
  {$IFDEF WINCE}
  WINSOCK2_DLL = 'ws2.dll';
  {$ELSE}
  WINSOCK2_DLL = 'ws2_32.dll';
  {$ENDIF}

const
  {$IFDEF FPC}
  INVALID_SOCKET: TSocket = $FFFFFFFF;
  {$ENDIF}
  CF_ACCEPT = $0000;
  CF_REJECT = $0001;
  SO_CONDITIONAL_ACCEPT = $3002;

type
  TSignature = array [0..5] of Char;
  TBlockType = (btNull, btData, btPing, btPong, btInfo);
  TBlockHeader = packed record
    Signature: TSignature;
    BlockType: TBlockType;
    DataSize: Integer;
  end;
  PBlockHeader = ^TBlockHeader;

const
  Signature: TSignature = 'BlSck';

function WSAAccept(S: TSocket; Addr: PSockAddr; AddrLen: PInteger;
                   lpfnCondition: Pointer; dwCallbackData: Integer): TSocket; stdcall; external WINSOCK2_DLL;

////////////////////////////////////////////////////////////////////////////

const
  CM_SOCKETMESSAGE = WM_USER;


procedure TCustomSocket.SetBlockMode(const Value: Boolean);
begin
  if FActive then
    raise ESocketException.Create('Cannot change this property in active state')
  else
    FBlockMode:= Value;
end;

function TCustomSocket.GetSendBufferLength: Integer;
begin
  if FActive then
    Result:= FSendBuffer.Size
  else
    Result:= 0;
end;

function TCustomSocket.FindBlockBegin(const DataArray: Pointer;
  const DataSize: Integer): Pointer;
var
  i: Integer;
begin
  Result:= nil;
  if (DataArray <> nil) and (DataSize >= SizeOf(TSignature)) then
    begin
      i:= 0;
      while i <= DataSize-SizeOf(TSignature) do
        begin
          if TSignature(Pointer(Integer(DataArray)+i)^) = Signature then
            begin
              Result:= Pointer(Integer(DataArray)+i);
              i:= DataSize;
            end
          else
            inc(i);
        end;
    end;
end;

procedure TCustomSocket.OnDataReceived(const Data: Pointer; const Size: Integer);
var
  BlockBegin, Buff: Pointer;
  Header: TBlockHeader;
begin
  if Size > 0 then
    if FBlockMode then
      begin
        ReallocMem(FBlockBuffer.Data, FBlockBuffer.Size+Size);
        Move(Data^, Pointer(Integer(FBlockBuffer.Data)+FBlockBuffer.Size)^, Size);
        Inc(FBlockBuffer.Size, Size);
        BlockBegin:= FindBlockBegin(FBlockBuffer.Data, FBlockBuffer.Size);
        if (BlockBegin = nil) and (FBlockBuffer.Size >= SizeOf(TSignature)) then
          begin
            MoveMemory(FBlockBuffer.Data, Pointer(Integer(FBlockBuffer.Data)+FBlockBuffer.Size-(SizeOf(TSignature)-1)), SizeOf(TSignature)-1);
            FBlockBuffer.Size:= SizeOf(TSignature)-1;
            ReallocMem(FBlockBuffer.Data, FBlockBuffer.Size);
          end;
        while BlockBegin <> nil do
          begin
            if Integer(BlockBegin) <= Integer(FBlockBuffer.Data)+FBlockBuffer.Size-SizeOf(TBlockHeader) then
              if TBlockHeader(BlockBegin^).DataSize > FMaxBlockSize then
                Inc(Integer(BlockBegin)) // - 
              else
                if TBlockHeader(BlockBegin^).DataSize <= FBlockBuffer.Size-(Integer(BlockBegin)-Integer(FBlockBuffer.Data))-SizeOf(TBlockHeader) then
                  begin
                    Header:= TBlockHeader(BlockBegin^);
                    GetMem(Buff, Header.DataSize);
                    Move(Pointer(Integer(BlockBegin)+SizeOf(TBlockHeader))^, Buff^, Header.DataSize);

                    Dec(FBlockBuffer.Size, Integer(BlockBegin)-Integer(FBlockBuffer.Data)+SizeOf(TBlockHeader)+Header.DataSize);
                    MoveMemory(FBlockBuffer.Data, Pointer(Integer(BlockBegin)+SizeOf(TBlockHeader)+Header.DataSize), FBlockBuffer.Size);
                    ReallocMem(FBlockBuffer.Data, FBlockBuffer.Size);
                    BlockBegin:= FBlockBuffer.Data;
                    case Header.BlockType of
                      btNull: OnNullPacketReceived;
                      btData:
                        if Assigned(FOnIncoming) then
                          try
                            FOnIncoming(Self, Buff, Header.DataSize);
                          except
                            { nothing }
                          end;
                      btPing: if Header.DataSize = SizeOf(Cardinal) then Pong(PCardinal(Buff)^);
                      btPong:
                        if (Header.DataSize = SizeOf(Cardinal)) and Assigned(FOnPong) then
                          try
                            FOnPong(Self, GetTickCount-PCardinal(Buff)^);
                          except
                            // nothing
                          end;
                      btInfo: SendInfo;    
                    end;
                    FreeMem(Buff, Header.DataSize);
                  end
                else
                  BlockBegin:= nil
            else
              Inc(Integer(BlockBegin));
            if Integer(BlockBegin) > Integer(FBlockBuffer.Data) then BlockBegin:= FindBlockBegin(BlockBegin, FBlockBuffer.Size-(Integer(BlockBegin)-Integer(FBlockBuffer.Data)));
          end;
      end
    else
      if Assigned(FOnIncoming) then
        try
          FOnIncoming(Self, Data, Size);
        except
          { nothing }
        end;
end;

procedure TCustomSocket.OnNullPacketReceived;
begin
  // nothing, only for overriding
end;

procedure TCustomSocket.ProcessSendBuffer;
var
  Sent: Integer;
begin
  if FSendBuffer.Size > 0 then
    try
      Sent:= TrySend(FSendBuffer.Data, FSendBuffer.Size);
      if (Sent > 0) and (Sent < FSendBuffer.Size) then
        MoveMemory(FSendBuffer.Data, Pointer(Integer(FSendBuffer.Data)+Sent), FSendBuffer.Size-Sent);
      Dec(FSendBuffer.Size, Sent);
      ReallocMem(FSendBuffer.Data, FSendBuffer.Size);
    except
      // nothing
    end;
end;

function TCustomSocket.TrySend(const Data: Pointer; const Size: Integer): Integer;
var
  Res, Err: Integer;
begin
  Result:= 0;
  Err:= 0;
  repeat
    Res:= WinSock.send(FSocket, Pointer(Integer(Data)+Result)^, Min(FMaxWSAPacketSize, Size-Result), 0);
    if Res <> SOCKET_ERROR then
      Inc(Result, Res)
    else
      begin
        Err:= WSAGetLastError;
        if Err <> WSAEWOULDBLOCK then RaiseError(seWrite, Err);
      end;
  until (Err <> 0) or (Result = Size);
end;

procedure TCustomSocket.SafeSend(const Data: Pointer; const Size: Integer);
var
  Sent: Integer;
begin
  if FSendBuffer.Size > 0 then
    begin
      ReallocMem(FSendBuffer.Data, FSendBuffer.Size+Size);
      Move(Data^, Pointer(Integer(FSendBuffer.Data)+FSendBuffer.Size)^, Size);
      Inc(FSendBuffer.Size, Size);
    end
  else
    begin
      Sent:= TrySend(Data, Size);
      if Sent < Size then
        begin
          FSendBuffer.Size:= Size-Sent;
          GetMem(FSendBuffer.Data, FSendBuffer.Size);
          Move(Pointer(Integer(Data)+Sent)^, FSendBuffer.Data^, FSendBuffer.Size);
        end;
    end;
end;

function TCustomSocket.Send(const Data: Pointer; const Size: Integer): Boolean;
var
  Header: TBlockHeader;
begin
  Result:= false;
  if FActive and (Size <= FMaxBlockSize) and (Size > 0) then
    begin
      FSocketLock.Enter;
      try
        if FBlockMode then
          begin
            Header.Signature:= Signature;
            Header.BlockType:= btData;
            Header.DataSize:= Size;
            SafeSend(@Header, SizeOf(Header));
          end;
        SafeSend(Data, Size);
        Result:= true;
      finally
        FSocketLock.Leave;
      end;
    end
  else
    WSASetLastError(WSAENOTSOCK);

  if not Result then RaiseError(seWrite, WSAGetLastError);
end;

function TCustomSocket.Ping: Boolean;
var
  Data: Pointer;
  Size: Integer;
begin
  if FBlockMode then
    try
      Size:= SizeOf(TBlockHeader)+SizeOf(Cardinal);
      GetMem(Data, Size);
      PBlockHeader(Data)^.Signature:= Signature;
      PBlockHeader(Data)^.BlockType:= btPing;
      PBlockHeader(Data)^.DataSize:= SizeOf(Cardinal);
      PCardinal(Integer(Data)+SizeOf(TBlockHeader))^:= GetTickCount;
      FSocketLock.Enter;
      try
        SafeSend(Data, Size);
      finally
        FSocketLock.Leave;
        FreeMem(Data, Size);
      end;
      Result:= true;
    except
      Result:= false;
    end
  else
    raise ESocketException.Create('This method allowed only if BlockMode = TRUE');
end;

function TCustomSocket.Pong(const PingParam: Cardinal): Boolean;
var
  Data: Pointer;
  Size: Integer;
begin
  try
    Size:= SizeOf(TBlockHeader)+SizeOf(Cardinal);
    GetMem(Data, Size);
    PBlockHeader(Data)^.Signature:= Signature;
    PBlockHeader(Data)^.BlockType:= btPong;
    PBlockHeader(Data)^.DataSize:= SizeOf(PingParam);
    PCardinal(Integer(Data)+SizeOf(TBlockHeader))^:= PingParam;
    FSocketLock.Enter;
    try
      SafeSend(Data, Size);
    finally
      FSocketLock.Leave;
      FreeMem(Data, Size);
    end;
    Result:= true;
  except
    Result:= false;
  end;
end;

function TCustomSocket.SendNullPacket: Boolean;
var
  Header: TBlockHeader;
begin
  if FBlockMode then
    try
      Header.Signature:= Signature;
      Header.BlockType:= btNull;
      Header.DataSize:= 0;
      FSocketLock.Enter;
      try
        SafeSend(@Header, SizeOf(Header));
      finally
        FSocketLock.Leave;
      end;
      Result:= true;
    except
      Result:= false;
    end
  else
    raise ESocketException.Create('This method allowed only if BlockMode = TRUE');
end;

function TCustomSocket.QueryInfo: Boolean;
var
  Header: TBlockHeader;
begin
  if FBlockMode then
    try
      Header.Signature:= Signature;
      Header.BlockType:= btInfo;
      Header.DataSize:= 0;
      FSocketLock.Enter;
      try
        SafeSend(@Header, SizeOf(Header));
      finally
        FSocketLock.Leave;
      end;
      Result:= true;
    except
      Result:= false;
    end
  else
    raise ESocketException.Create('This method allowed only if BlockMode = TRUE');
end;

function TCustomSocket.SendInfo: Boolean;
var
  Data: Pointer;
  Size: Integer;
begin
  try
    Size:= SizeOf(TBlockHeader)+SizeOf(TSimpleSocketInfo);
    GetMem(Data, Size);
    PBlockHeader(Data)^.Signature:= Signature;
    PBlockHeader(Data)^.BlockType:= btData;
    PBlockHeader(Data)^.DataSize:= SizeOf(TSimpleSocketInfo);
    PSimpleSocketInfo(Integer(Data)+SizeOf(TBlockHeader))^.Size:= SizeOf(TSimpleSocketInfo);
    PSimpleSocketInfo(Integer(Data)+SizeOf(TBlockHeader))^.Version:= CurVersion;
    PSimpleSocketInfo(Integer(Data)+SizeOf(TBlockHeader))^.Comment:= 'Misha Cherkes aka MIKE (C) 2007';
    FSocketLock.Enter;
    try
      SafeSend(Data, Size);
    finally
      FSocketLock.Leave;
      FreeMem(Data, Size);
    end;
    Result:= true;
  except
    Result:= false;
  end;
end;

function TCustomSocket.WaitFor(const Action: TWaitAction; const TimeOut: Cardinal = 0): Boolean;
var
  EndTime: Cardinal;
begin
  EndTime:= GetTickCount+TimeOut;
  case Action of
    waOpen:
      begin
        while (not FActive) and (FSocket <> INVALID_SOCKET) and ((TimeOut = 0) or (GetTickCount < EndTime)) do
          begin
            Sleep(10);
            ProcessMessages;
          end;
        Result:= FActive;
      end;
    waSend:
      begin
        while FActive and (FSendBuffer.Size > 0) and ((TimeOut = 0) or (GetTickCount < EndTime)) do
          begin
            Sleep(10);
            ProcessMessages;
          end;
        Result:= FActive and (FSendBuffer.Size = 0);
      end;
  else
    Result:= false;    
  end;
end;

constructor TCustomSocket.Create;
var
  WSALData: WSAData;
begin
  inherited Create;
  FActive:= false;
  FSocket:= INVALID_SOCKET;
  FAddress.sin_family:= PF_INET;
  FAddress.sin_addr.S_addr:= INADDR_ANY;
  FAddress.sin_port:= 0;
  FMaxWSAPacketSize:= 4096;
  FMaxBlockSize:= 10*1024*1024;
  FOnIncoming:= nil;
  FOnError:= nil;
  FOnPong:= nil;
  FBlockMode:= false;
  FTag:= 0;
  FSocketLock:= TCriticalSection.Create;
  FSocketLock.Enter;
  WSAStartup($0101, WSALData);
  FSocketLock.Leave;
end;

destructor TCustomSocket.Destroy;
begin
  FSocketLock.Enter;
  WSACleanup;
  FSocketLock.Leave;
  FSocketLock.Free;
  Close;
  inherited Destroy;
end;

procedure TCustomSocket.Open;
begin
  if not FActive then
    begin
      FBlockBuffer.Data:= nil;
      FBlockBuffer.Size:= 0;
      FSendBuffer.Data:= nil;
      FSendBuffer.Size:= 0;
      FActive:= true;
    end;
end;

procedure TCustomSocket.Close;
begin
  if FActive then
    begin
      ReallocMem(FBlockBuffer.Data, 0);
      ReallocMem(FSendBuffer.Data, 0);
      FActive:= false;
    end;
end;

function TCustomSocket.Min(const Arg1, Arg2: Integer): Integer;
begin
  if Arg1 < Arg2 then
    Result:= Arg1
  else
    Result:= Arg2;
end;

procedure TCustomSocket.RaiseError(const Error: TSocketErrorType; const Code: Integer);
var
  Handled: Boolean;
begin
  Handled:= false;
  if Assigned(FOnError) then
    try
      FOnError(Self, Error, Code, Handled);
    except
      { nothing }
    end;
  if not Handled then
    raise ESocketException.CreateResFmt(@sWindowsSocketError, [SysErrorMessage(Code), Code, SocketErrorsText[Error]]);
end;

procedure TCustomSocket.DataReceive;
var
  Size, Received: Integer;
  Buf: Pointer;
begin
  Buf:= nil;
  if ioctlsocket(FSocket, FIONREAD, Size) = 0 then
    if Size > 0 then
      try
        getmem(Buf, Size);
        FSocketLock.Enter;
        Received:= recv(FSocket, Buf^, Size, 0);
        FSocketLock.Leave;
        if Received <> SOCKET_ERROR then
          OnDataReceived(Buf, Received)
        else
          RaiseError(seRead, WSAGetLastError);
      finally
        if Buf <> nil then freemem(Buf, Size);
      end;
end;

procedure TCustomSocket.SetActive(const Value: Boolean);
begin
  if Value then
    Open
  else
    Close;
end;

procedure TCustomSocket.SetMaxBlockSize(const Value: Integer);
const
  MinValue = 1024;
begin
  if Value < MinValue then
    FMaxBlockSize:= MinValue
  else
    FMaxBlockSize:= Value;
end;

procedure TCustomSocket.SetMaxWSAPacketSize(const Value: Word);
const
  MinValue = 16;
begin
  if Value < MinValue then
    FMaxWSAPacketSize:= MinValue
  else
    FMaxWSAPacketSize:= Value;
end;

////////////////////////////////////////////////////////////////////////////

constructor TSimpleClientSocket.Create;
begin
  inherited Create;
  FWndHandle:= INVALID_HANDLE_VALUE;
  FPort:= 0;
  FHost:= '';
  FOnConnect:= nil;
  FOnDisconnect:= nil;
end;

destructor TSimpleClientSocket.Destroy;
begin
  Close;
  inherited Destroy;
end;

procedure TSimpleClientSocket.FreeResources;
var
  LastError: Integer;
begin
  LastError:= WSAGetLastError;
  try
    if FSocket <> INVALID_SOCKET then
      try
        FSocketLock.Enter;
        if closesocket(FSocket) <> 0 then RaiseError(seClose, WSAGetLastError);
      finally
        FSocketLock.Leave;
        FSocket:= INVALID_SOCKET;
        Sleep(10);
      end;
  finally
    if FWndHandle <> INVALID_HANDLE_VALUE then
      try
        ProcessMessages;
      finally
        DeallocateHWnd(FWndHandle);
        FWndHandle:= INVALID_HANDLE_VALUE;
        WSASetLastError(LastError);
      end;
  end;
end;

function TSimpleClientSocket.LookupName(const Name: String): TInAddr;
var
  HostEnt: PHostEnt;
  InAddr: TInAddr;
begin
  HostEnt:= gethostbyname(PChar(Name));
  FillChar(InAddr, SizeOf(InAddr), 0);
  if HostEnt <> nil then
    begin
      InAddr.S_un_b.s_b1:= HostEnt^.h_addr^[0];
      InAddr.S_un_b.s_b2:= HostEnt^.h_addr^[1];
      InAddr.S_un_b.s_b3:= HostEnt^.h_addr^[2];
      InAddr.S_un_b.s_b4:= HostEnt^.h_addr^[3];
    end;
  Result:= InAddr;
end;

procedure TSimpleClientSocket.Open;
var
  Res: Integer;
begin
  if not FActive then
    begin
      FWndHandle:= AllocateHwnd(WindowProc);
      if FWndHandle <> INVALID_HANDLE_VALUE then
        begin
          FSocket:= WinSock.socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
          if FSocket <> INVALID_SOCKET then
            begin
              FAddress.sin_family:= PF_INET;
              FAddress.sin_port:= htons(FPort);
              FAddress.sin_addr.S_addr:= inet_addr(PChar(FHost));
              if FAddress.sin_addr.S_addr = INVALID_SOCKET then FAddress.sin_addr:= LookupName(FHost);
              if WSAAsyncSelect(FSocket, FWndHandle, CM_SOCKETMESSAGE, FD_READ or FD_WRITE or FD_CONNECT or FD_CLOSE) = 0 then
                if connect(FSocket, FAddress, SizeOf(FAddress)) <> 0 then
                  begin
                    Res:= WSAGetLastError;
                    if Res <> WSAEWOULDBLOCK then
                      begin
                        FreeResources;
                        RaiseError(seOpen, Res);
                      end;
                  end
                else
                  Sleep(10)
              else
                begin
                  FreeResources;
                  RaiseError(seOpen, WSAGetLastError);
                end;
            end
          else
            begin
              FreeResources;
              RaiseError(seOpen, WSAGetLastError);
            end;
        end
      else
        begin
          FreeResources;
          RaiseError(seOpen, WSAGetLastError);
        end;
    end;
end;


procedure TSimpleClientSocket.Close;
begin
  if FActive then
    begin
      FreeResources;
      inherited Close;
      if Assigned(FOnDisconnect) then
        try
          FOnDisconnect(Self);
        except
          { nothing }
        end;
    end;
end;

procedure TSimpleClientSocket.ProcessMessages;
var
  Msg: TMsg;
begin
  while PeekMessage(Msg, FWndHandle, 0, 0, PM_REMOVE) do
    begin
      TranslateMessage(Msg);
      DispatchMessage(Msg);
    end;
end;

procedure TSimpleClientSocket.WindowProc(var Msg: TMessage);
begin
  if Msg.Msg = CM_SOCKETMESSAGE then
    begin
      Msg.Result:= 0;
      if Msg.LParamHi = 0 then // 
        case Msg.LParamLo of
          FD_READ: DataReceive;
          FD_WRITE: ProcessSendBuffer;
          FD_CONNECT: begin
                        inherited Open;
                        if Assigned(FOnConnect) then
                          try
                            FOnConnect(Self);
                          except
                            { nothing }
                          end;
                      end;
          FD_CLOSE: Close;
        else dispatch(Msg);
        end
      else // 
        case Msg.LParamLo of
          FD_READ: RaiseError(seRead, Msg.LParamHi);
          FD_WRITE:
            if Msg.LParamHi <> WSAEWOULDBLOCK then
              RaiseError(seWrite, Msg.LParamHi);
          FD_CONNECT:
            if Msg.LParamHi <> WSAEWOULDBLOCK then
               begin
                 FActive:= false;
                 FreeResources;
                 RaiseError(seOpen, Msg.LParamHi);
               end;
          FD_CLOSE: begin
                      if FActive then //           ,   
                        Close
                      else
                        RaiseError(seClose, Msg.LParamHi);
                    end;
        else RaiseError(seGeneral, Msg.LParamHi);
        end;
    end
  else
    dispatch(Msg);
end;

procedure TSimpleClientSocket.OnQueryEndSession(var Msg: TMessage);
begin
  Msg.Result:= Integer(true); //   
end;

procedure TSimpleClientSocket.SetHost(const Value: String);
begin
  if not FActive then FHost:= Value;
end;

procedure TSimpleClientSocket.SetPort(const Value: Word);
begin
  if not FActive then FPort:= Value;
end;

////////////////////////////////////////////////////////////////////////////

constructor TServerClientSocket.Create(const AServerSocket: TSimpleServerSocket);
begin
  inherited Create;
  FServerSocket:= AServerSocket;
  FOnAcceptRequest:= nil;
  FLocalAddress.sin_family:= PF_INET;
  FLocalAddress.sin_addr.S_addr:= INADDR_ANY;
  FLocalAddress.sin_port:= 0;
end;

destructor TServerClientSocket.Destroy;
begin
  Close;
  inherited Destroy;
end;

procedure TServerClientSocket.FreeResources;
var
  LastError: Integer;
begin
  LastError:= WSAGetLastError;
  if FSocket <> INVALID_SOCKET then
    try
      FSocketLock.Enter;
      if closesocket(FSocket) <> 0 then RaiseError(seClose, WSAGetLastError);
    finally
      FSocketLock.Leave;
      FSocket:= INVALID_SOCKET;
      WSASetLastError(LastError);
    end;
end;

//   -  TServerClientSocket.Open   FreePascal !!!
function AcceptCallBack(lpCallerId, lpCallerData: PWSABuf; lpSQOS, lpGQOS: PQOS;
                        lpCalleeID, lpCalleeData: PWSABuf; g: PGroup;
                        dwCallbackData: Integer): Integer; stdcall;
var
  Allow: Boolean;
  ServerClientSocket: TServerClientSocket;
begin
  ServerClientSocket:= TServerClientSocket(dwCallbackData);

  if Assigned(lpCallerId) and (lpCallerId^.Len = SizeOf(TSockAddr)) then
    ServerClientSocket.FAddress:= PSockAddr(lpCallerId^.Buf)^;
  if Assigned(lpCalleeId) and (lpCalleeId^.Len = SizeOf(TSockAddr)) then
    ServerClientSocket.FLocalAddress:= PSockAddr(lpCalleeId^.Buf)^;

  Allow:= true;
  if Assigned(ServerClientSocket) and Assigned(ServerClientSocket.OnAcceptRequest) then
    try
      ServerClientSocket.OnAcceptRequest(ServerClientSocket, Allow);
    except
      // nothing
    end;

  if Allow then
    Result:= CF_ACCEPT
  else
    Result:= CF_REJECT;
end;

procedure TServerClientSocket.Open;
var
  Len: Integer;
begin
  if Assigned(FServerSocket) and (not FActive) then
    begin
      Len:= SizeOf(FAddress);
      FSocket:= WSAAccept(FServerSocket.Socket, @FAddress, @Len, @AcceptCallBack, Integer(Self));
      if FSocket <> INVALID_SOCKET then
        if WSAAsyncSelect(FSocket, FServerSocket.WndHandle, CM_SOCKETMESSAGE, FD_READ or FD_WRITE or FD_CLOSE) = 0 then
          begin
            if FLocalAddress.sin_addr.S_addr = 0 then
              begin
                Len:= SizeOf(FLocalAddress);
                getsockname(FSocket, FLocalAddress, LongInt(Len));
              end;
            inherited Open;
            Sleep(10);
          end
        else
          begin
            FreeResources;
            RaiseError(seAccept, WSAGetLastError);
          end
      else
        if WSAGetLastError <> WSAEConnRefused then
          begin
            FreeResources;
            RaiseError(seAccept, WSAGetLastError);
          end;
    end;
end;

procedure TServerClientSocket.Close;
begin
  if FActive then
    begin
      FreeResources;
      inherited Close;
    end;
end;

procedure TServerClientSocket.ProcessMessages;
begin
  FServerSocket.ProcessMessages;
end;

procedure TServerClientSocket.Drop;
begin
  FServerSocket.SendDropMessage(FSocket);
end;

////////////////////////////////////////////////////////////////////////////

constructor TSimpleServerSocket.Create;
var
  WSALData: WSAData;
begin
  inherited Create;
  WSAStartup($0101, WSALData);
  FWndHandle:= INVALID_HANDLE_VALUE;
  FActive:= false;
  FSocket:= INVALID_SOCKET;
  FAddress.sin_family:= PF_INET;
  FAddress.sin_addr.S_addr:= INADDR_ANY;
  FAddress.sin_port:= 0;
  FPort:= 0;
  FMaxWSAPacketSize:= 4096;
  FMaxBlockSize:= 10*1024*1024;
  FBlockMode:= false;
  FTag:= 0;
  FOnIncoming:= nil;
  FOnOpen:= nil;
  FOnClose:= nil;
  FOnClientConnected:= nil;
  FOnClientDisconnected:= nil;
  FOnAcceptRequest:= nil;
  FOnClientError:= nil;
  FOnClientPong:= nil;
  FSocketLock:= TCriticalSection.Create;
  FClientSockets:= TIndexList.Create;
end;

destructor TSimpleServerSocket.Destroy;
begin
  Close;
  FClientSockets.Free;
  FSocketLock.Free;
  WSACleanup;
  inherited Destroy;
end;

procedure TSimpleServerSocket.RaiseError(const Sender: TObject; const Error: TSocketErrorType; const Code: Integer);
var
  Handled: Boolean;
begin
  Handled:= false;
  if Assigned(FOnError) then
    try
      FOnError(Sender, Error, Code, Handled);
    except
      { nothing }
    end;
  if not Handled then
    raise ESocketException.CreateResFmt(@sWindowsSocketError, [SysErrorMessage(Code), Code, SocketErrorsText[Error]]);
end;

procedure TSimpleServerSocket.FreeResources;
var
  i, LastError: Integer;
begin
  LastError:= WSAGetLastError;
  try
    if FSocket <> INVALID_SOCKET then
      try
        FSocketLock.Enter;
        try
          i:= FClientSockets.FirstIndex;
          while i <> -1 do
            begin
              SendDropMessage(TServerClientSocket(FClientSockets[i]).Socket);
              i:= FClientSockets.NextIndex(i);
            end;
        finally
          if closesocket(FSocket) <> 0 then RaiseError(Self, seClose, WSAGetLastError);
        end;
      finally
        FSocketLock.Leave;
        FSocket:= INVALID_SOCKET;
        Sleep(10);
      end;
  finally
    if FWndHandle <> INVALID_HANDLE_VALUE then
      try
        ProcessMessages; // !!!
      finally
        DeallocateHWnd(FWndHandle);
        FWndHandle:= INVALID_HANDLE_VALUE;
        WSASetLastError(LastError);
      end;
  end;
end;

procedure TSimpleServerSocket.Open;
var
  Value: Bool;
begin
  if not FActive then
    begin
      FWndHandle:= AllocateHwnd(WindowProc);
      if FWndHandle <> INVALID_HANDLE_VALUE then
        begin
          FSocket:= WinSock.socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
          if FSocket <> INVALID_SOCKET then
            begin
              Value:= true;
              SetSockOpt(FSocket, SOL_SOCKET, SO_CONDITIONAL_ACCEPT, @Value, SizeOf(Value));

              FAddress.sin_family:= PF_INET;
              FAddress.sin_port:= htons(FPort);
              FAddress.sin_addr.s_addr:= INADDR_ANY;
              if bind(FSocket, FAddress, SizeOf(FAddress)) <> 0 then
                begin
                  FreeResources;
                  RaiseError(Self, seOpen, WSAGetLastError);
                end
              else
                if (WSAAsyncSelect(FSocket, FWndHandle, CM_SOCKETMESSAGE, FD_ACCEPT) = 0) and (Winsock.listen(FSocket, SOMAXCONN) = 0) then
                  begin
                    FActive:= true;
                    Sleep(10);
                    if Assigned(FOnOpen) then
                      try
                        FOnOpen(Self);
                      except
                        { nothing }
                      end;
                  end
                else
                  begin
                    FreeResources;
                    RaiseError(Self, seOpen, WSAGetLastError);
                  end;
            end
          else
            begin
              FreeResources;
              RaiseError(Self, seOpen, WSAGetLastError);
            end;
        end
      else
        begin
          FreeResources;
          RaiseError(Self, seOpen, 0);
        end;
    end;
end;

procedure TSimpleServerSocket.Close;
begin
  if FActive then
    begin
      FreeResources;
      FActive:= false;
      if Assigned(FOnClose) then
        try
          FOnClose(Self);
        except
          { nothing }
        end;
    end;
end;

procedure TSimpleServerSocket.ProcessMessages;
var
  Msg: TMsg;
begin
  while PeekMessage(Msg, FWndHandle, 0, 0, PM_REMOVE) do
    begin
      TranslateMessage(Msg);
      DispatchMessage(Msg);
    end;
end;

function TSimpleServerSocket.FindClientSocket(const SocketToFind: TSocket): Integer;
var
  i: Integer;
begin
  Result:= -1;
  if (SocketToFind <> INVALID_SOCKET) and (FClientSockets.Count > 0) then
    begin
      FSocketLock.Enter;
      i:= FClientSockets.FirstIndex;
      while i <> -1 do
        if (FClientSockets[i] <> nil) and (TServerClientSocket(FClientSockets[i]).Socket = SocketToFind) then
          begin
            Result:= i;
            i:= -1;
          end
        else
          i:= FClientSockets.NextIndex(i);
      FSocketLock.Leave;
    end;
end;

procedure TSimpleServerSocket.WindowProc(var Msg: TMessage);
var
  Index: Integer;
begin
  if Msg.Msg = CM_SOCKETMESSAGE then
    begin
      Msg.Result:= 0;
      if Msg.LParamHi = 0 then // 
        case Msg.LParamLo of
          FD_READ: begin
                     Index:= FindClientSocket(TSocket(Msg.WParam));
                     if Index <> -1 then TServerClientSocket(FClientSockets[Index]).DataReceive;
                   end;
          FD_WRITE: begin
                      Index:= FindClientSocket(TSocket(Msg.WParam));
                      if Index <> -1 then TServerClientSocket(FClientSockets[Index]).ProcessSendBuffer;
                    end;
          FD_ACCEPT: begin
                       FSocketLock.Enter;
                       Index:= FClientSockets.Add(TServerClientSocket.Create(Self));
                       FSocketLock.Leave;
                       TServerClientSocket(FClientSockets[Index]).MaxWSAPacketSize:= FMaxWSAPacketSize;
                       TServerClientSocket(FClientSockets[Index]).BlockMode:= FBlockMode;
                       TServerClientSocket(FClientSockets[Index]).MaxBlockSize:= FMaxBlockSize;
                       TServerClientSocket(FClientSockets[Index]).OnIncoming:= OnClientEntryIncoming;
                       TServerClientSocket(FClientSockets[Index]).OnError:= OnClientEntryError;
                       TServerClientSocket(FClientSockets[Index]).OnAcceptRequest:= OnClientAcceptRequest;
                       TServerClientSocket(FClientSockets[Index]).OnPong:= OnClientEntryPong;
                       TServerClientSocket(FClientSockets[Index]).Open;
                       if TServerClientSocket(FClientSockets[Index]).Active then
                         begin
                           if Assigned(FOnClientConnected) then
                             try
                               FOnClientConnected(Self, TServerClientSocket(FClientSockets[Index]));
                             except
                               { nothing }
                             end;
                         end
                       else
                         try
                           FSocketLock.Enter;
                           TServerClientSocket(FClientSockets[Index]).Free;
                         finally
                           FClientSockets.Delete(Index);
                           FSocketLock.Leave;
                         end;
                     end;
          FD_CLOSE: begin
                      Index:= FindClientSocket(TSocket(Msg.WParam));
                      if Index <> -1 then
                        begin
                          if Assigned(FOnClientDisconnected) then
                            try
                              FOnClientDisconnected(Self, TServerClientSocket(FClientSockets[Index]));
                            except
                              { nothing }
                            end;
                          try
                            FSocketLock.Enter;
                            TServerClientSocket(FClientSockets[Index]).Free;
                          finally
                            FClientSockets.Delete(Index);
                            FSocketLock.Leave;
                          end;
                        end;
                    end;
        else dispatch(Msg);
        end
      else // 
        if TSocket(Msg.WParam) = FSocket then
          case Msg.LParamLo of
            FD_READ: RaiseError(Self, seRead, Msg.LParamHi);
            FD_WRITE: RaiseError(Self, seWrite, Msg.LParamHi);
            FD_ACCEPT: RaiseError(Self, seAccept, Msg.LParamHi);
            FD_CLOSE: RaiseError(Self, seClose, Msg.LParamHi);
          else RaiseError(Self, seGeneral, Msg.LParamHi);
          end
        else
          begin
            Index:= FindClientSocket(TSocket(Msg.WParam));
            if Index <> -1 then
              case Msg.LParamLo of
                FD_READ: RaiseError(TServerClientSocket(FClientSockets[Index]), seRead, Msg.LParamHi);
                FD_WRITE:
                  if Msg.LParamHi <> WSAEWOULDBLOCK then
                    RaiseError(TServerClientSocket(FClientSockets[Index]), seWrite, Msg.LParamHi);
                FD_ACCEPT: RaiseError(TServerClientSocket(FClientSockets[Index]), seAccept, Msg.LParamHi);
                FD_CONNECT: RaiseError(TServerClientSocket(FClientSockets[Index]), seOpen, Msg.LParamHi);
                FD_CLOSE: begin
                            if Assigned(FOnClientDisconnected) then
                              try
                                FOnClientDisconnected(Self, TServerClientSocket(FClientSockets[Index]));
                              except
                                { nothing }
                              end;
                            try
                              FSocketLock.Enter;
                              TServerClientSocket(FClientSockets[Index]).Free;
                            finally
                              FClientSockets.Delete(Index);
                              FSocketLock.Leave;
                            end;
                          end;
              else RaiseError(TServerClientSocket(FClientSockets[Index]), seGeneral, Msg.LParamHi);
              end;
          end;
    end
  else
    Dispatch(Msg);
end;

procedure TSimpleServerSocket.OnQueryEndSession(var Msg: TMessage);
begin
  Msg.Result:= Integer(true); //     
end;

procedure TSimpleServerSocket.SetBlockMode(const Value: Boolean);
begin
  if FActive then
    raise ESocketException.Create('Cannot change this property in active state')
  else
    FBlockMode:= Value;
end;

procedure TSimpleServerSocket.SetActive(const Value: Boolean);
begin
  if Value then
    Open
  else
    Close;
end;

procedure TSimpleServerSocket.SetPort(const Value: Word);
begin
  if not FActive then FPort:= Value;
end;

function TSimpleServerSocket.GetClient(const Index: Integer): TServerClientSocket;
begin
  FSocketLock.Enter;
  Result:= FClientSockets[Index];
  FSocketLock.Leave;
end;

function TSimpleServerSocket.GetClientCount: Integer;
begin
  FSocketLock.Enter;
  Result:= FClientSockets.Count;
  FSocketLock.Leave;
end;

function TSimpleServerSocket.GetClientBySocket(const SocketToFind: TSocket): TServerClientSocket;
var
  Index: Integer;
begin
  Index:= FindClientSocket(SocketToFind);
  if Index = -1 then
    Result:= nil
  else
    begin
      FSocketLock.Enter;
      Result:= FClientSockets[Index];
      FSocketLock.Leave;
    end;
end;

function TSimpleServerSocket.GetSendBufferLength: Integer;
var
  i: Integer;
begin
  Result:= 0;
  FSocketLock.Enter;
  try
    i:= GetNextClientIndex(-1);
    while i >= 0 do
      begin
        Inc(Result, Clients[i].SendBufferLength);
        i:= GetNextClientIndex(i);
      end;
  finally
    FSocketLock.Leave;
  end;
end;

procedure TSimpleServerSocket.SetMaxBlockSize(const Value: Integer);
const
  MinValue = 1024;
begin
  if Value < MinValue then
    FMaxBlockSize:= MinValue
  else
    FMaxBlockSize:= Value;
end;

procedure TSimpleServerSocket.SetMaxWSAPacketSize(const Value: Word);
const
  MinValue = 16;
begin
  if Value < MinValue then
    FMaxWSAPacketSize:= MinValue
  else
    FMaxWSAPacketSize:= Value;
end;

procedure TSimpleServerSocket.SendDropMessage(const Socket: TSocket);
begin
  PostMessage(FWndHandle, CM_SOCKETMESSAGE, Socket, FD_CLOSE and $0000FFFF);
end;

procedure TSimpleServerSocket.DropClientSocket(const Index: Integer);
begin
  FSocketLock.Enter;
  if FClientSockets.IndexExists(Index) then
    SendDropMessage(TServerClientSocket(FClientSockets[Index]).Socket);
  FSocketLock.Leave;
end;

function TSimpleServerSocket.GetNextClientIndex(
  const CurIndex: Integer): Integer;
begin
  if CurIndex < 0 then
    Result:= FClientSockets.FirstIndex
  else
    Result:= FClientSockets.NextIndex(CurIndex);
end;

procedure TSimpleServerSocket.SendNullPacketAll;
var
  i: Integer;
begin
  if FBlockMode then
    begin
      FSocketLock.Enter;
      i:= FClientSockets.FirstIndex;
      while i >= 0 do
        begin
          TServerClientSocket(FClientSockets[i]).SendNullPacket;
          i:= FClientSockets.NextIndex(i);
        end;
      FSocketLock.Leave;
    end
  else
    raise ESocketException.Create('This method allowed only if BlockMode = TRUE');
end;

procedure TSimpleServerSocket.PingAll;
var
  i: Integer;
begin
  if FBlockMode then
    begin
      FSocketLock.Enter;
      i:= FClientSockets.FirstIndex;
      while i >= 0 do
        begin
          TServerClientSocket(FClientSockets[i]).Ping;
          i:= FClientSockets.NextIndex(i);
        end;
      FSocketLock.Leave;
    end
  else
    raise ESocketException.Create('This method allowed only if BlockMode = TRUE');
end;

procedure TSimpleServerSocket.OnClientEntryError(Sender: TObject; const Error: TSocketErrorType; const Code: Integer; var Handled: Boolean);
begin
  if Assigned(FOnClientError) then
    try
      FOnClientError(Self, Sender as TServerClientSocket, Error, Code, Handled);
    except
      // nothing
    end;
end;

procedure TSimpleServerSocket.OnClientEntryIncoming(Sender: TObject; const Data: Pointer; const Size: Integer);
begin
  if Assigned(FOnIncoming) then
    try
      FOnIncoming(Self, Sender as TServerClientSocket, Data, Size);
    except
      // nothing
    end;
end;

procedure TSimpleServerSocket.OnClientAcceptRequest(Sender: TObject;
  var AllowAccept: Boolean);
begin
  if Assigned(FOnAcceptRequest) then
    try
      FOnAcceptRequest(Self, Sender as TServerClientSocket, AllowAccept);
    except
      // nothing
    end;
end;

procedure TSimpleServerSocket.OnClientEntryPong(Sender: TObject; const PingTime: Cardinal);
begin
  if Assigned(FOnClientPong) then
    try
      FOnClientPong(Self, Sender as TServerClientSocket, PingTime);
    except
      // nothing
    end;
end;

////////////////////////////////////////////////////////////////////////////

function SockAddrToIP(const Address: TSockAddr): TIP;
begin
  Result.Long:= Address.sin_addr.S_addr;
end;

function IPToStr(const IP: TIP): String;
begin
  Result:= IntToStr(IP.b1)+'.'+IntToStr(IP.b2)+'.'+IntToStr(IP.b3)+'.'+IntToStr(IP.b4);
end;

function StrToIP(const Address: String): TIP;
var
  s: String;
  i, c: Integer;
begin
  i:= 1;
  for c:= 1 to 4 do
    begin
      s:= '';
      while (i <= Length(Address)) and (Address[i] <> '.') do
        begin
          s:= s+Address[i];
          Inc(i);
        end;
      Result.Bytes[c]:= StrToInt(s);
      Inc(i);
    end;
end;


end.


